from torchhd.datasets.beijing_air_quality import BeijingAirQuality
from torchhd.datasets.isolet import ISOLET
from torchhd.datasets.european_languages import EuropeanLanguages
from torchhd.datasets.ucihar import UCIHAR
from torchhd.datasets.airfoil_self_noise import AirfoilSelfNoise
from torchhd.datasets.emg_hand_gestures import EMGHandGestures
from torchhd.datasets.pamap import PAMAP
from torchhd.datasets.ccpp import CyclePowerPlant
from torchhd.datasets.dataset import CollectionDataset
from torchhd.datasets.dataset import DatasetFourFold
from torchhd.datasets.dataset import DatasetTrainTest
from torchhd.datasets.abalone import Abalone
from torchhd.datasets.adult import Adult


__all__ = [
    "BeijingAirQuality",
    "ISOLET",
    "EuropeanLanguages",
    "UCIHAR",
    "AirfoilSelfNoise",
    "EMGHandGestures",
    "PAMAP",
    "CyclePowerPlant",
    "CollectionDataset",
    "DatasetFourFold",
    "DatasetTrainTest",
    "Abalone",
    "Adult",
]
